package Anarres::Mud::Driver::Program::Node;

use strict;
use vars qw(@ISA @EXPORT_OK %EXPORT_TAGS @NODETYPES $BASEPACK);
use Carp qw(:DEFAULT cluck);
use Exporter;
use Data::Dumper;
use Anarres::Mud::Driver::Compiler::Lex;
use Anarres::Mud::Driver::Program::Type qw(:types);
use Anarres::Mud::Driver::Program::Efun qw(%EFUNS);

sub CONST		()	{ 1 }
sub LVALUE		()	{ 2 }
sub PURE		()	{ 4 }

{
	my $i = 0;
	sub tempname	()	{ return 'tmp' . $i++; }
}

q{

	Allowed promotions:

	int => bool (defined and nonzero)
	(all others just definedness)
	array => bool
	mapping => bool
	object => bool
	string => bool (definedness)
	closure => bool
	nil => bool (false)

	int => string (for addition, hash deref, assignment)
	object => string (for hash deref)

	XXX Type promotion should be a method on the type object.

} if 0;

BEGIN {
	my @FLAGS = qw(CONST LVALUE PURE);
	@ISA = qw(Exporter);
	@EXPORT_OK = (@FLAGS, qw(error t_integer t_string));
	%EXPORT_TAGS = (
		flags	=> \@FLAGS,
		all		=> \@EXPORT_OK,
			);

	# Vivify the relevant packages

	$BASEPACK = __PACKAGE__;	# XXX Use this instead of a literal
	@NODETYPES = qw(
		Assert
		Nil String Integer Array Mapping Closure
		Variable Parameter Funcall CallOther
		IndexArray IndexMap Range
		Catch Sscanf
		Postinc Postdec Preinc Predec Unot Tilde Uplus Uminus
		Lsh Rsh Dot Plus Minus Mult Divide Mod
		ArrAdd ArrSub ArrUnion ArrIsect
		MapAdd
		IntEq IntNe IntLt IntGt IntLe IntGe
		StrEq StrNe StrLt StrGt StrLe StrGe
		LogOr LogAnd BitOr BitAnd BitXor
		CondExp Assign Block StmtExp
		StmtDo StmtWhile StmtFor StmtForeach
		StmtRlimits StmtTry
		StmtIf StmtSwitch StmtCase
		StmtBreak StmtContinue StmtReturn
			);

	my $PACKAGE = __PACKAGE__;
	foreach (@NODETYPES) {
		eval qq{
			package $PACKAGE\::$_;
			use strict;
			use vars qw(\@ISA);
			use Data::Dumper;
			use Anarres::Mud::Driver::Program::Node qw(:all);
			use Anarres::Mud::Driver::Program::Type qw(:types);
			use Anarres::Mud::Driver::Compiler qw(\$PROGRAM);
			use Anarres::Mud::Driver::Program::Efun qw(%EFUNS);
			\@ISA = qw(Anarres::Mud::Driver::Program::Node);
			# *EFUNS = \\\%Anarres::Mud::Driver::Program::Efun::EFUNS;
		}; die $@ if $@;
	}
}

# In the Node package...

sub new {
	my ($class, $type, $flags, @vals) = @_;
	confess "Construct node!" if $class eq 'node';
	# print "Construct $class: $type with " . scalar(@vals) . " vals\n";
	unless ($type) {
		(my $name = $class) =~ s/^.*:://;
		confess "Construct typeless $name node\n";
	}
	my $self = [ $type, $flags, @vals ];
	return bless $self, $class;
}

sub error { die @_ }

# By default, statements are assumed void.
sub infer {
	my ($self, $newtype) = @_;
	if (ref($self) =~ /Stmt/) {
		print "$_[0] Infers only as void\n";
		return $_[1]->equals(T_VOID) ? $_[0] : undef
	}
	my $type = $self->type;
	# print "Inferring $type to $newtype\n";
	return $type->promote($self, $newtype);
}

sub type	{ $_[0]->[0] }
sub flags	{ $_[0]->[1] }
sub value	{ $_[0]->[2 + $_[1]] }
sub values	{
	my $self = shift;
	# print "Values are length " . (scalar(@$self) - 2) . "\n";
	my @vals = @$self;
	return @vals[2..$#vals];
	# return $self->[2..$#$self]
}
sub setflag	{ $_[0]->[1] |= $_[1] }

sub assert {
	my ($self, $type) = @_;
	return new Anarres::Mud::Driver::Program::Node::Assert($type,$self);
}

# my $infer_assert = sub { $_[0]->assert($_[1]) }
# { ... *infer = $infer_assert; } etc
my $infer_integer = sub { $_[1]->equals(T_INTEGER) ? $_[0] : undef };
my $infer_string = sub { $_[1]->equals(T_STRING) ? $_[0] : undef };

my %OPTABLE = (
	Postinc	=> 'A ++',
	Postdec	=> 'A --',
	Preinc	=> '++ A',
	Predec	=> '-- A',
	Unot	=> '! A',
	Tilde	=> '~ A',
	Uplus	=> '+ A',
	Uminus	=> '- A',

	Lsh		=> 'A << B',
	Rsh		=> 'A >> B',
	Dot		=> 'A . B',
	Plus	=> 'A + B',
	Minus	=> 'A - B',
	Mult	=> 'A * B',
	Divide	=> 'A / B',
	Mod		=> 'A % B',

	IntEq	=> 'A == B',
	IntNe	=> 'A != B',
	IntLt	=> 'A < B',
	IntGt	=> 'A > B',
	IntLe	=> 'A <= B',
	IntGe	=> 'A >= B',

	StrEq	=> 'A eq B',
	StrNe	=> 'A ne B',
	StrLt	=> 'A lt B',
	StrGt	=> 'A gt B',
	StrLe	=> 'A le B',
	StrGe	=> 'A ge B',

	LogOr	=> 'A || B',
	LogAnd	=> 'A && B',
	BitOr	=> 'A | B',
	BitAnd	=> 'A & B',
	BitXor	=> 'A ^ B',

	CondExp		=> 'A ? B : C',

	IndexArray	=> 'A->[B]',
	IndexMap	=> 'A->{B}',
	Range		=> '[ A->[B..C] ]',

	ArrAdd		=> '[ @{A}, @{B} ]',
	MapAdd		=> '{ %{A}, %{B} ]',

	Assign		=> 'A = B',
	Catch		=> '(eval { A; }, $@)',

	StmtReturn	=> 'return A;',
	StmtBreak	=> 'break;',
	StmtContinue=> 'continue;',

	# We can add extra braces around statement|block tokens
	# This lot are all strictly cheating anyway! If this works ...
	StmtDo		=> 'do { A } while (B);',
	StmtWhile	=> 'while (A) { B }',
	StmtFor		=> 'for (A; B; C) { D }',
	StmtForeach	=> 'foreach (@{ A }) { B }',
	StmtTry		=> 'eval A; if ($@) B;',		# This uses blocks
		);

# "Refactor", I hear you say?
sub generate {
	my $self = shift;

	my $name = ref($self);
	$name =~ s/^.*:://;						# A package name
	my $code = $OPTABLE{$name} or return "GEN($name)";

	foreach ('A'..'D') {	# Say ...
		my $arg = ord($_) - ord('A');
		$code =~ s/$_/' . \$self->value($arg)->generate(\@_) . '/;
	}

	$code = qq{ sub { my \$self = shift; return '$code'; } };
	# Remove empty concatenations - careful with the templates
	$code =~ s/'' \. //g;
	$code =~ s/ \. ''//g;

	print "$name becomes $code\n";
	my $subref = eval $code;
	die $@ if $@;

	{
		# Backpatch our original package.
		no strict qw(refs);
		*{ ref($self) . '::generate' } = $subref;
	}

	return $subref->($self, @_);
}

sub dump {
	my $self = shift;

	my $op = ref($self);
	$op =~ s/.*:://;

	my @vals = $self->values;
	my @rest = map { ! $_ ? "[undef]" :
					! ref($_) ? $_ :
					ref($_) !~ /::/ ? "$_" :
					$_->dump(@_) } $self->values;
	unshift(@rest, lc $op);

	return "([" . $self->type->dump(@_) . "] " . join(" ", @rest) . ")";
}

sub t_integer {	# Modifies its arguments
	my $base = $_[0];
	my $func = $_[1];
	foreach (@_[2..$#_]) {
		$_ = $_->infer(T_INTEGER);
		unless ($_) {
			error "Argument $base to $func not integer";
		}
		$base++;
	}
}

sub t_string {	# Modifies its arguments
	my $base = $_[0];
	my $func = $_[1];
	foreach (@_[2..$#_]) {
		# print "t_string($_)\n";
		confess "t_string on nonref" unless ref $_;
		$_ = $_->infer(T_STRING);
		unless ($_) {
			error "Argument $base to $func not string";
		}
		$base++;
	}
}

{
	package Anarres::Mud::Driver::Program::Node::Assert;
	sub new { return __PACKAGE__->SUPER::new($_[1], 0, $_[2]); }
	# sub infer { $_[1]->equals($_[0]->type) ? $_[0] : undef; }
	sub generate {
		my $self = shift;
		# Check the type at runtime
		"\n# XXX Check type at runtime\n" .
		$self->value(0)->generate(@_);
	}
}

{
	package Anarres::Mud::Driver::Program::Node::Nil;
	sub new { return __PACKAGE__->SUPER::new(T_NIL, CONST); }
	# sub infer { $_[0]; }	# Always true
	sub generate { "undef"; }
}

{
	package Anarres::Mud::Driver::Program::Node::String;
	use String::Escape qw(quote printable);
	sub new { return __PACKAGE__->SUPER::new(T_STRING, CONST, $_[1]); }
	# sub infer { $_[1]->equals(T_STRING) ? $_[0] : undef }
	sub dump { quote(printable($_[0]->value(0))) }
	sub generate { quote(printable($_[0]->value(0))) }
}

{
	package Anarres::Mud::Driver::Program::Node::Integer;
	sub new { return __PACKAGE__->SUPER::new(T_INTEGER, CONST, $_[1]); }
	sub infer {
		if ($_[1]->equals(T_INTEGER)) {
			return $_[0];
		}
		elsif ($_[0]->value(0) == 0) {	# A valid nil
			return new Anarres::Mud::Driver::Program::Node::Nil;
		}
		elsif ($_[1]->equals(T_STRING)) {
			# XXX This wants a generic promote routine.
			# This should really use $program->method('to_string')
			return new Anarres::Mud::Driver::Program::Node::Funcall(
							$EFUNS{"to_string"},	# XXX Naughty
							[ $_[0] ],
								);
		}
		else {
			return undef;
		}
	}

	sub dump { $_[0]->value(0) }
	sub generate { $_[0]->value(0) }
}

{
	package Anarres::Mud::Driver::Program::Node::Array;
	sub new {
		my (undef, $vals) = @_;
		my $flag = CONST;
		my $type = 0;
		ARRAYCONST: {
			foreach (@$vals) {
				# XXX Search the types as well to find a good type.
				$type = $_->type->meet($type);
				$flag &= $_->flags;
			}
		}
		$type ||= T_ARRAY;
		return __PACKAGE__->SUPER::new($type->pointer, $flag, @$vals);
	}
	sub old_infer {
		return undef unless $_[1]->arrayp;
		# XXX continue for full type. Are we an array of type X?
		return $_[0];
	}
	sub generate {
		my ($self, $indent, @rest) = @_;
		$indent++;

		my @vals = map { $_->generate($indent, @rest) } $self->values;

		$indent--;
		my $isep = "\n" . ("\t" x $indent);
		my $sep = "," . $isep . "\t";
		return "[" . $isep . "\t" . join($sep, @vals) . $isep . "]";
	}
}

{
	package Anarres::Mud::Driver::Program::Node::Mapping;
	sub new {
		my (undef, $vals) = @_;
		# XXX Check left hand sides infer to string.
		my $self = __PACKAGE__->SUPER::new(T_MAPPING, 0, @$vals);
		MAPPINGCONST: {
			foreach (@$vals) {
				last MAPPINGCONST unless $_->flags & CONST;
			}
			$self->setflag(CONST);
		}
		return $self;
	}
	sub old_infer { $_[1]->equals(T_MAPPING) ? $_[0] : undef }
	sub generate {
		my ($self, $indent, @rest) = @_;
		$indent++;

		my @vals = map { $_->generate($indent, @rest) } $self->values;
		my @out = ();
		
		while (my @tmp = splice(@vals, 0, 2)) {
			push(@out, $tmp[0] . "\t=> " . $tmp[1]);
		}

		$indent--;
		my $isep = "\n" . ("\t" x $indent);
		my $sep = "," . $isep . "\t";
		return "{$isep\t" . join($sep, @out) . ",$isep}";
	}
}

{
	package Anarres::Mud::Driver::Program::Node::Closure;
	sub new { shift; return __PACKAGE__->SUPER::new(T_CLOSURE, CONST, @_); }
	sub old_infer { $_[1]->equals(T_CLOSURE) ? $_[0] : undef }
	# XXX This needs to store the owner object so we can emulate the
	# LPC behaviour of function_owner. Something like [ $self, sub {} ]
	sub generate {
		my $self = shift;
		return "sub { " . $self->value(0)->generate(@_) . " }";
	}
}

{
	package Anarres::Mud::Driver::Program::Node::Variable;
	# XXX Look up type
	sub new {
		my $var = $_[1];	 # A variable object, or undef
		error "No such variable $_[2]" unless $var;
		return __PACKAGE__->SUPER::new($var->type, LVALUE, $_[1]);
	}
	# XXX As an rvalue? Delegate to a basic type infer method.
	# XXX If it's an rvalue then it must be initialised. Also for ++, --
	sub old_infer { $_[1]->equals($_[0]->type) ? $_[0] : undef }
	sub dump {
		my $self = shift;
		my $var = $self->value(0);
		"([" . $var->type->dump(@_) . "] variable '" . $var->name . "')"
	}
	sub generate {
		# XXX This is WRONG if it's a global!
		return '$' . $_[0]->value(0)->name;
	}
}

{
	package Anarres::Mud::Driver::Program::Node::Parameter;
	# XXX Look up type
	sub new { shift; return __PACKAGE__->SUPER::new(T_MIXED, LVALUE, @_); }
	# XXX We could look this up at the current point ...
	sub infer { $_[0]->assert($_[1]) }	# XXX As an rvalue?
	sub generate { '$_[' . $_[0]->value(0) . ']' }
}

{
	package Anarres::Mud::Driver::Program::Node::Funcall;
	# XXX Look up return type, number of args
	sub new {
		my (undef, $method, $args) = @_;
		error "Method not found" unless $method;
		return __PACKAGE__->SUPER::new($method->type, 0,$method,@$args);
	}
	# XXX Or assert
	sub old_infer { $_[1]->equals($_[0]->type) ? $_[0] : undef }
	sub dump {
		my $self = shift;
		my @args = $self->values;
		my $method = shift @args;
		"([" . $method->type->dump(@_) . "] funcall '" . $method->name .
						"' " . scalar(@args) . " args)"
	}
	sub generate {
		my $self = shift;
		my @args = $self->values;
		my $method = shift @args;
		@args = map { $_->generate(@_) } @args;
		return $method->gencall(@args);
	}
}

{
	package Anarres::Mud::Driver::Program::Node::CallOther;
	# XXX Look up return type
	sub new {
		my (undef, $exp, $name, $args) = @_;
		return __PACKAGE__->SUPER::new(T_MIXED, 0, $exp, $name, @$args);
	}
	sub infer { $_[0]->assert($_[1]) }
	sub dump {
		my $self = shift;
		my @args = $self->values;
		my $exp = shift @args;
		my $name = shift @args;
		"([_] callother " . $exp->dump(@_) . " -> '" . $name .
						"' " . scalar(@args) . " args)"
	}
	sub generate {
		my $self = shift;
		my @args = $self->values;
		my $exp = shift @args;
		my $name = shift @args;
		@args = map { $_->generate(@_) } @args;
		return '(' . $exp->generate(@_) . ')->' . $name . '(' .
						join(", ", @args) . ')';
						
	}
}

# I could call this 'Index' and call IndexMap 'Lookup'
{
	package Anarres::Mud::Driver::Program::Node::IndexArray;
	# XXX Look up return type
	sub new {
		my (undef, $arr, $idx) = @_;

		my $at = $arr->type;
		my $it = $idx->type;

		my $temp;

		my $flags = $arr->flags & $idx->flags & CONST;
		$flags |= LVALUE if $arr->flags & LVALUE;

		if ($at->arrayp) {
			$idx = $idx->infer(T_INTEGER);
			error "Array index not integer" unless $idx;
			return __PACKAGE__->SUPER::new(
							$at->deref, $flags, $arr, $idx);
		}
		elsif ($at->equals(T_MAPPING)) {
			# XXX Allow promotion of object?
			$idx = $idx->infer(T_STRING);
			error "Mapping index not string" unless $idx;
			return new Anarres::Mud::Driver::Program::Node::IndexMap(
							T_MIXED, $flags, $arr, $idx);
		}
		elsif ($temp = $arr->infer(T_ARRAY)) {
			$idx = $idx->infer(T_INTEGER);
			error "Array index not integer" unless $idx;
			return __PACKAGE__->SUPER::new(
							$at->deref, $flags, $temp, $idx);
		}
		elsif ($temp = $arr->infer(T_MAPPING)) {
			# XXX Allow promotion of object?
			$idx = $idx->infer(T_STRING);
			error "Mapping index not string" unless $idx;
			return new Anarres::Mud::Driver::Program::Node::IndexMap(
							T_MIXED, $flags, $temp, $idx);
		}
		else {
			error "Can only index mapping or array, not " .
			$arr->type->dump;
		}
	}
	sub old_infer { $_[0]->assert($_[1]) }
	sub old_generate {
		my $self = shift;
		return $self->value(0)->generate(@_) . '->[' .
				 $self->value(1)->generate(@_) . ']';
	}
}

{
	package Anarres::Mud::Driver::Program::Node::IndexMap;
	sub infer { $_[0]->assert($_[1]) }
	sub old_generate {
		my $self = shift;
		return $self->value(0)->generate(@_) . '->{' .
				 $self->value(1)->generate(@_) . '}';
	}
}

{
	package Anarres::Mud::Driver::Program::Node::Range;
	sub new {
		my (undef, $arr, $start, $end) = @_;

		print "Infer on " . ref($arr) . "\n";

		error "Range may only be applied to array"
			unless $arr = $arr->infer(T_ARRAY);
		error "Start of range must be integer"
			unless $start = $start->infer(T_INTEGER);
		error "End of range must be integer"
			unless $end = $end->infer(T_INTEGER);

		return __PACKAGE__->SUPER::new($arr->type, 0,
						$arr, $start, $end);
	}
	sub old_infer {
		# XXX Copy Array::infer
		return undef unless $_[1]->arrayp;
		return $_[0];
	}
}

{
	package Anarres::Mud::Driver::Program::Node::Catch;
	sub new { shift; return __PACKAGE__->SUPER::new(T_STRING, 0, @_); }
	# sub infer { $_[0]->assert($_[1]) }
}

{
	package Anarres::Mud::Driver::Program::Node::Sscanf;
	# This should be $_[1], @{$_[2]}
	sub new { shift; return __PACKAGE__->SUPER::new(T_INTEGER, 0, @_); }
	# sub infer { $_[0]->assert($_[1]) }
}

foreach ( qw(Postinc Postdec Preinc Predec) ) {
	my $n = lc $_;
	eval qq{
		{
			package Anarres::Mud::Driver::Program::Node::$_;
			sub new {
				my \$val = \$_[1]->infer(T_INTEGER);
				error "Argument to $n not integer" unless \$val;
				error "Argument to $n not lvalue"
								unless \$val->flags & LVALUE;
				return __PACKAGE__->SUPER::new(T_INTEGER, 0, \$val);
			}
			# *infer = \$infer_integer;
		}
	}; die $@ if $@;
}

{
	package Anarres::Mud::Driver::Program::Node::Unot;
	sub new {
		my $val = $_[1];

		my $vt = $val->type;

		if ($vt->equals(T_NIL)) {
			warn "Expression always true due to type";
			return new Anarres::Mud::Driver::Program::Node::Integer(1);
		}
		elsif ($val = $val->infer(T_INTEGER)) {
			return __PACKAGE__->SUPER::new(T_INTEGER, 0, $val);
		}
		else {
			warn "Expression always false due to type";
			return new Anarres::Mud::Driver::Program::Node::Nil;
		}
	}
	# *infer = $infer_integer;
}

foreach ( qw(Tilde Uplus Uminus) ) {
	eval qq{
		{
			package Anarres::Mud::Driver::Program::Node::$_;
			sub new { shift;
				t_integer(0, lc "$_", \@_);
				return __PACKAGE__->SUPER::new(T_INTEGER, 0, \@_);
			}
			# *infer = \$infer_integer;
		}
	}; die $@ if $@;
}

foreach ( qw(Lsh Rsh Mult Divide Mod)) {
	eval qq{
		{
			package Anarres::Mud::Driver::Program::Node::$_;
			sub new { shift;
				t_integer(0, "integer " . lc "$_", \@_);
				return __PACKAGE__->SUPER::new(T_INTEGER, 0, \@_);
			}
			# *infer = \$infer_integer;
		}
	}; die $@ if $@;
}

{
	package Anarres::Mud::Driver::Program::Node::Dot;
	sub new { shift;
		t_string(0, "concatenation", @_);
		return __PACKAGE__->SUPER::new(T_STRING, 0, @_);
	}
	# *infer = $infer_string;
}

{
	package Anarres::Mud::Driver::Program::Node::Plus;
	sub new {
		my (undef, $left, $right) = @_;

		my $li = $left->infer(T_INTEGER);
		my $ri = $right->infer(T_INTEGER);
		if ($li && $ri) {
			# XXX Pass const flags
			return __PACKAGE__->SUPER::new(T_INTEGER, 0, $li, $ri);
		}

		my $ls = $left->infer(T_STRING);
		my $rs = $right->infer(T_STRING);
		if ($ls && $rs) {
			return new Anarres::Mud::Driver::Program::Node::Dot(
							$ls, $rs,
								);
		}

		my $la = $left->infer(T_ARRAY);
		my $ra = $right->infer(T_ARRAY);
		if ($la && $ra) {
			return new Anarres::Mud::Driver::Program::Node::ArrAdd(
							$la, $ra,
								);
		}

		my $lm = $left->infer(T_MAPPING);
		my $rm = $right->infer(T_MAPPING);
		if ($lm && $rm) {
			return new Anarres::Mud::Driver::Program::Node::MapAdd(
							$lm, $rm,
								);
		}

		error("Cannot add types " . $left->type->dump . " and " .
						$right->type->dump);
	}
	# *infer = $infer_integer;	# The only native case
}

{
	package Anarres::Mud::Driver::Program::Node::Minus;
	sub new {
		my (undef, $left, $right) = @_;

		my $li = $left->infer(T_INTEGER);
		my $ri = $right->infer(T_INTEGER);
		if ($li && $ri) {
			# XXX Pass const flags
			return __PACKAGE__->SUPER::new(T_INTEGER, 0, $li, $ri);
		}

		# No string case

		my $la = $left->infer(T_ARRAY);
		my $ra = $left->infer(T_ARRAY);
		if ($la && $ra) {
			return new Anarres::Mud::Driver::Program::Node::ArrSub(
							$la, $ra,
								);
		}

		error("Cannot subtract types " . $left->type->dump . " and " .
						$right->type->dump);
	}
	# *infer = $infer_integer;	# The only native case
}

{
	package Anarres::Mud::Driver::Program::Node::ArrAdd;
	# sub infer { $_[1]->arrayp ? $_[0] : undef }
}

{
	package Anarres::Mud::Driver::Program::Node::MapAdd;
	# sub infer { $_[1]->equals(T_MAPPING) ? $_[0] : undef }
}

{
	package Anarres::Mud::Driver::Program::Node::ArrSub;
	# sub infer { $_[1]->arrayp ? $_[0] : undef }
	sub invoke {	# Can't really do this inline?
		my %table = map { $_ => 1 } @{ $_[1] };
		return [ grep { ! $table{$_} } @{ $_[0] } ]
	}
	sub generate {
		my $self = shift;
		return 'Anarres::Mud::Driver::Program::Node::ArrSub::invoke(' .
						$self->value(0)->generate(@_) . ', ' .
						$self->value(1)->generate(@_) . ')';
	}
}

{
	package Anarres::Mud::Driver::Program::Node::ArrUnion;
	# sub infer { $_[1]->arrayp ? $_[0] : undef }
	sub invoke {
		my @left = @{ $_[0] };
		my %table = map { $_ => 1 } @left;
		foreach (@{ $_[1] }) {
			push(@left, $_) unless $table{$_};
		}
		return \@left;
	}
	sub generate {
		my $self = shift;
		return 'Anarres::Mud::Driver::Program::Node::ArrUnion::invoke('.
						$self->value(0)->generate(@_) . ', ' .
						$self->value(1)->generate(@_) . ')';
	}
}

{
	package Anarres::Mud::Driver::Program::Node::ArrIsect;
	# sub infer { $_[1]->arrayp ? $_[0] : undef }
	sub invoke {
		my @out = ();
		my %table = map { $_ => 1 } @{ $_[1] };
		foreach (@{ $_[0] }) {
			push(@out, $_) if $table{$_};
		}
		return \@out;
	}
	sub generate {
		my $self = shift;
		return 'Anarres::Mud::Driver::Program::Node::ArrIsect::invoke('.
						$self->value(0)->generate(@_) . ', ' .
						$self->value(1)->generate(@_) . ')';
	}
}

foreach (qw(IntEq IntNe IntLt IntGt IntLe IntGe)) {
	my $strop = $_;
	$strop =~ s/^Int/Str/;
	eval qq{
		{
			package Anarres::Mud::Driver::Program::Node::$_;
			sub new {
				my (undef, \$left, \$right) = \@_;

				if (\$left->type->equals(T_STRING) ||
					\$right->type->equals(T_STRING)) {
					t_string(0, "string " . lc "$_", \$left, \$right);
					return new
							Anarres::Mud::Driver::Program::Node::$strop(
									T_STRING, 0, \$left, \$right,
										);
				}
				# This can serve for anything except strings.
				return __PACKAGE__->SUPER::new(
								T_INTEGER, 0, \$left, \$right,
									);
			}
			# *infer = \$infer_integer;
		}
		{
			package Anarres::Mud::Driver::Program::Node::$strop;
			# new called correctly from above hence no override needed
			# *infer = \$infer_string;
		}
	}; die $@ if $@;
}

{
	package Anarres::Mud::Driver::Program::Node::LogOr;
	sub new { shift;
		return __PACKAGE__->SUPER::new(T_INTEGER, 0, @_);
	}
	# *infer = $infer_integer;
}

{
	package Anarres::Mud::Driver::Program::Node::LogAnd;
	sub new { shift;
		return __PACKAGE__->SUPER::new(T_INTEGER, 0, @_);
	}
	# *infer = $infer_integer;
}

foreach (qw(BitOr BitAnd BitXor)) {
	eval qq{
		{
			package Anarres::Mud::Driver::Program::Node::$_;
			sub new { shift;
				t_integer(0, "integer " . lc "$_", \@_);
				return __PACKAGE__->SUPER::new(T_INTEGER, 0, \@_);
			}
			# *infer = \$infer_integer;
		}
	}; die $@ if $@;
}

{
	package Anarres::Mud::Driver::Program::Node::CondExp;
	sub new {
		my (undef, $cond, $true, $false) = @_;
		# XXX Check types and const $cond
		return __PACKAGE__->SUPER::new($true->type, 0,
						$cond, $true, $false);
	}
	# *infer = $infer_string;
}

{
	package Anarres::Mud::Driver::Program::Node::Assign;
	sub new {
		my (undef, $var, $exp) = @_;
		# XXX Check types
		# Set initialised flag
		my $val = $exp->infer($var->type);
		die
			"Type mismatch in assignment: " .
				$var->type->dump . " = " . $exp->type->dump
					unless $val;
		return __PACKAGE__->SUPER::new($var->type, 0, $var, $val);
	}
	# *infer = $infer_string;
}

{
	package Anarres::Mud::Driver::Program::Node::Block;
	sub new { shift; return __PACKAGE__->SUPER::new(T_VOID, 0, @_); }
	sub dump {
		my ($self, $indent) = @_;
		$indent++;

		my @locals = map { $_->dump($indent, @_) } @{ $self->value(0) };
		my @stmts = map { $_->dump($indent, @_) } @{ $self->value(1) };
		my $sep = "\n" . ("\t" x $indent);
		return join($sep, "([V] block", @locals, @stmts) . ")";
	}
	sub generate {
		my ($self, $indent, @rest) = @_;
		$indent++;

		my @args = map { $_->name } @{ $self->value(0) };
		my $args = @args
						? 'my ($' . join(', $', @args) . ');'
						: '# no locals in block';
		my @vals = map { $_->generate($indent, @rest) }
						@{ $self->value(1) };

		$indent--;
		my $isep = "\n" . ("\t" x $indent);
		my $sep = $isep . "\t";
		return $isep . join($sep, "{", $args, @vals) . $isep . "}";
	}
}

{
	package Anarres::Mud::Driver::Program::Node::StmtExp;
	sub new { shift; return __PACKAGE__->SUPER::new(T_VOID, 0, @_); }
	sub generate {
		my $self = shift;
		return $self->value(0)->generate(@_) . ';';
	}
}

{
	package Anarres::Mud::Driver::Program::Node::StmtRlimits;
		# XXX Check types
	sub new { shift; return __PACKAGE__->SUPER::new(T_VOID, 0, @_); }
	sub generate {
		my $self = shift;
		return $self->value(3)->generate(@_) . ';';
	}
}

{
	package Anarres::Mud::Driver::Program::Node::StmtIf;
	# XXX Promote $_[1] to bool
	sub new { shift; return __PACKAGE__->SUPER::new(T_VOID, 0, @_); }
	sub generate {
		my ($self, $indent, @args) = @_;
		my $sep = "\t" x $indent;
		$indent++;
		my $out =
			"if (" .
				$self->value(0)->generate($indent, @args) . ") {\n" .
				$sep . "\t" .
					$self->value(1)->generate($indent, @args) . "\n" .
			$sep . "}";
		my $else = $self->value(2);
		$out .=
			"\n" . $sep . "else {\n" .
			$sep . "\t" .
				$else->generate($indent, @args) . "\n" .
			$sep . "}";
		return $out;
	}
}

foreach (qw(
		StmtDo StmtWhile StmtFor StmtForeach
		StmtTry
		StmtSwitch StmtCase
		StmtBreak StmtContinue
		StmtReturn
		)) {
	eval qq{
		{
			package Anarres::Mud::Driver::Program::Node::$_;
			sub new { shift;
				return __PACKAGE__->SUPER::new(T_VOID, 0, \@_);
			}
		}
	}; die $@ if $@;
}

{
	no strict qw(refs);
	my @nogen;
	foreach (@NODETYPES) {
		next if defined $OPTABLE{$_};
		next if defined &{ "$BASEPACK\::$_\::generate" };
		push(@nogen, $_);
	}
	print "No generate in @nogen\n";
}

1;
